/*
 *
 * Function for recognizing partitionability
 * (for pure 3-dimensional complexes)
 *
 * by M.Hachimori, Jul 2006.
 */

#include <stdio.h>
#include <stdlib.h>
#include "s_complex.h"
/* #define MAXVERT 100 */

int m_empty;
int m_vert[MAXVERT];
int m_edge[MAXVERT][MAXVERT];
int m_tri[MAXVERT][MAXVERT][MAXVERT];
int c_empty;
int c_vert[MAXVERT];
int c_edge[MAXVERT][MAXVERT];
int c_tri[MAXVERT][MAXVERT][MAXVERT];
int f_vec[MAXDIM], h_vec[MAXDIM];

void cancel3(Simplices* facets, int num);
int check3(Simplices* facets, int num);
void putpart3(Simplices* facets, int num);
void printdata3(Simplices* facets);

/*
 * Function for recognizing partitionability
 * (for pure 3-dimensional complexes)
 *   Input: list of facets
 *   return 1 if partitionable
 *   return 0 if not partitionable
 */
int is_partitionable3(facets)
  Simplices *facets;
{

  Simplices *faces;
  Simplices *tmp_facets;
  int i, j, k, dim = 3;

  /* Check the complex is pure 3-dimensional.*/
  tmp_facets = facets;
  while(tmp_facets != NULL) {
    if (tmp_facets->vert[0] != 3) {
      printf("There are facets of dimension != 3. Error.\n");
      return(-1);
    }
    tmp_facets = tmp_facets->next;
  }

  /* get h-vector */
  faces = facets2faces(facets);
  get_fvector(faces, f_vec);
  f_to_h(f_vec, h_vec, facets->vert[0]);

  /* check h-vector is nonnegative */
  for (i = 0; i <= dim+1; i ++)
    if (h_vec[i] < 0) {
      printf("h-vector has negative component. Not partitionable\n");
      return(0);
    }

  /* initialization */
  /* c_xxx indicate whether faces are covered or not */
  for (i = 0; i < MAXVERT; i ++) {
    c_vert[i] = 0;
    for (j = 0; j < MAXVERT; j ++) {
      c_edge[i][j] = 0;
      for (k = 0; k < MAXVERT; k ++)
        c_tri[i][j][k] = 0;
      }
    }
  c_empty = 0;
  /* m_xxx indicate the multiplicity of incident facets */
  for (i = 0; i < MAXVERT; i ++) {
    m_vert[i] = 0;
    for (j = 0; j < MAXVERT; j ++) {
      m_edge[i][j] = 0; 
      for (k = 0; k < MAXVERT; k ++)
        m_tri[i][j][k] = 0;
      }
    }
  m_empty = 0;
  tmp_facets = facets;
  while(tmp_facets != NULL) {
    m_vert[tmp_facets->vert[1]] ++;
    m_vert[tmp_facets->vert[2]] ++;
    m_vert[tmp_facets->vert[3]] ++;
    m_vert[tmp_facets->vert[4]] ++;
    m_edge[tmp_facets->vert[1]][tmp_facets->vert[2]] ++;
    m_edge[tmp_facets->vert[1]][tmp_facets->vert[3]] ++;
    m_edge[tmp_facets->vert[1]][tmp_facets->vert[4]] ++;
    m_edge[tmp_facets->vert[2]][tmp_facets->vert[3]] ++;
    m_edge[tmp_facets->vert[2]][tmp_facets->vert[4]] ++;
    m_edge[tmp_facets->vert[3]][tmp_facets->vert[4]] ++;
    m_tri[tmp_facets->vert[1]][tmp_facets->vert[2]][tmp_facets->vert[3]] ++;
    m_tri[tmp_facets->vert[1]][tmp_facets->vert[2]][tmp_facets->vert[4]] ++;
    m_tri[tmp_facets->vert[1]][tmp_facets->vert[3]][tmp_facets->vert[4]] ++;
    m_tri[tmp_facets->vert[2]][tmp_facets->vert[3]][tmp_facets->vert[4]] ++;
    m_empty ++;
    tmp_facets->vert[5] = -1;  /* facets->vert[5] indicates the type of interval
                                 in a partition. 
                                   0: [empty,facet]
                                   1,2,3,4 : [1,f],[2,f],[3,f],[4,f]
                                   5,6,7,8,9,10 : [12,f],[13,f],[14,f],[23,f],[24,f],[34,f]
                                   11,12,13,14: [123,f],[124,f],[134,f],[234,f]
                                   15: [f,f]
                                   -1: not assigned yet.  */
    tmp_facets = tmp_facets->next;
  }

  /* main loop */
  tmp_facets = facets;
  while(tmp_facets != NULL) {
#ifdef DEBUG
    printdata3(facets);
#endif
    /* check types in the ordering -1,15,0,1,2,...,14 */
    if (tmp_facets->vert[5] == 14) {  /* backtrack */
      tmp_facets->vert[5] = -1;
      if (tmp_facets->prev == NULL) return 0;
      else {
        tmp_facets = tmp_facets->prev;
        cancel3(tmp_facets, tmp_facets->vert[5]);
        continue;
        }
      }
    else {
      if (tmp_facets->vert[5] == -1) tmp_facets->vert[5] = 15;
      else if (tmp_facets->vert[5] == 15) tmp_facets->vert[5] = 0;
      else tmp_facets->vert[5] ++;
      if (check3(tmp_facets, tmp_facets->vert[5])) {
        putpart3(tmp_facets, tmp_facets->vert[5]);
        if (tmp_facets->next == NULL) goto PARTITIONABLE;
        else {
          tmp_facets = tmp_facets->next;
          continue;
          }
        }
      }
    }

  return -1;

  PARTITIONABLE:
  printdata3(facets);
  return 1;

}

void cancel3(facets, num)
  Simplices* facets;
  int num;
{
  m_empty ++;
  m_vert[facets->vert[1]] ++;
  m_vert[facets->vert[2]] ++;
  m_vert[facets->vert[3]] ++;
  m_vert[facets->vert[4]] ++;
  m_edge[facets->vert[1]][facets->vert[2]] ++;
  m_edge[facets->vert[1]][facets->vert[3]] ++;
  m_edge[facets->vert[1]][facets->vert[4]] ++;
  m_edge[facets->vert[2]][facets->vert[3]] ++;
  m_edge[facets->vert[2]][facets->vert[4]] ++;
  m_edge[facets->vert[3]][facets->vert[4]] ++;
  m_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] ++;
  m_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] ++;
  m_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] ++;
  m_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] ++;
  switch(num) {
    case 0:
      c_empty --;
      c_vert[facets->vert[1]] --;
      c_vert[facets->vert[2]] --;
      c_vert[facets->vert[3]] --;
      c_vert[facets->vert[4]] --;
      c_edge[facets->vert[1]][facets->vert[2]] --;
      c_edge[facets->vert[1]][facets->vert[3]] --;
      c_edge[facets->vert[1]][facets->vert[4]] --;
      c_edge[facets->vert[2]][facets->vert[3]] --;
      c_edge[facets->vert[2]][facets->vert[4]] --;
      c_edge[facets->vert[3]][facets->vert[4]] --;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] --;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] --;
      c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] --;
      c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] --;
      h_vec[0] ++;
      break;
    case 1:
      c_vert[facets->vert[1]] --;
      c_edge[facets->vert[1]][facets->vert[2]] --;
      c_edge[facets->vert[1]][facets->vert[3]] --;
      c_edge[facets->vert[1]][facets->vert[4]] --;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] --;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] --;
      c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] --;
      h_vec[1] ++;
      break;
    case 2:
      c_vert[facets->vert[2]] --;
      c_edge[facets->vert[1]][facets->vert[2]] --;
      c_edge[facets->vert[2]][facets->vert[3]] --;
      c_edge[facets->vert[2]][facets->vert[4]] --;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] --;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] --;
      c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] --;
      h_vec[1] ++;
      break;
    case 3:
      c_vert[facets->vert[3]] --;
      c_edge[facets->vert[1]][facets->vert[3]] --;
      c_edge[facets->vert[2]][facets->vert[3]] --;
      c_edge[facets->vert[3]][facets->vert[4]] --;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] --;
      c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] --;
      c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] --;
      h_vec[1] ++;
      break;
    case 4:
      c_vert[facets->vert[4]] --;
      c_edge[facets->vert[1]][facets->vert[4]] --;
      c_edge[facets->vert[2]][facets->vert[4]] --;
      c_edge[facets->vert[3]][facets->vert[4]] --;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] --;
      c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] --;
      c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] --;
      h_vec[1] ++;
      break;
    case 5:
      c_edge[facets->vert[1]][facets->vert[2]] --;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] --;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] --;
      h_vec[2] ++;
      break;
    case 6:
      c_edge[facets->vert[1]][facets->vert[3]] --;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] --;
      c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] --;
      h_vec[2] ++;
      break;
    case 7:
      c_edge[facets->vert[1]][facets->vert[4]] --;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] --;
      c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] --;
      h_vec[2] ++;
      break;
    case 8:
      c_edge[facets->vert[2]][facets->vert[3]] --;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] --;
      c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] --;
      h_vec[2] ++;
      break;
    case 9:
      c_edge[facets->vert[2]][facets->vert[4]] --;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] --;
      c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] --;
      h_vec[2] ++;
      break;
    case 10:
      c_edge[facets->vert[3]][facets->vert[4]] --;
      c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] --;
      c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] --;
      h_vec[2] ++;
      break;
    case 11:
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] --;
      h_vec[3] ++;
      break;
    case 12:
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] --;
      h_vec[3] ++;
      break;
    case 13:
      c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] --;
      h_vec[3] ++;
      break;
    case 14:
      c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] --;
      h_vec[3] ++;
      break;
    case 15:
      break;
    }

  return;

}

int check3(facets, num)
  Simplices* facets;
  int num;
{

  /* h-vector check */
  switch(num) {
    case 0:
      if (h_vec[0] <= 0) {
#ifdef DEBUG
printf("h-vector fault\n");
#endif
        return 0;
        }
      break;
    case 1:
      if (h_vec[1] <= 0) {
#ifdef DEBUG
printf("h-vector fault\n");
#endif
        return 0;
        }
      break;
    case 2:
      if (h_vec[1] <= 0) {
#ifdef DEBUG
printf("h-vector fault\n");
#endif
        return 0;
        }
      break;
    case 3:
      if (h_vec[1] <= 0) {
#ifdef DEBUG
printf("h-vector fault\n");
#endif
        return 0;
        }
      break;
    case 4:
      if (h_vec[1] <= 0) {
#ifdef DEBUG
printf("h-vector fault\n");
#endif
        return 0;
        }
      break;
    case 5:
      if (h_vec[2] <= 0) {
#ifdef DEBUG
printf("h-vector fault\n");
#endif
        return 0;
        }
      break;
    case 6:
      if (h_vec[2] <= 0) {
#ifdef DEBUG
printf("h-vector fault\n");
#endif
        return 0;
        }
      break;
    case 7:
      if (h_vec[2] <= 0) {
#ifdef DEBUG
printf("h-vector fault\n");
#endif
        return 0;
        }
      break;
    case 8:
      if (h_vec[2] <= 0) {
#ifdef DEBUG
printf("h-vector fault\n");
#endif
        return 0;
        }
      break;
    case 9:
      if (h_vec[2] <= 0) {
#ifdef DEBUG
printf("h-vector fault\n");
#endif
        return 0;
        }
      break;
    case 10:
      if (h_vec[2] <= 0) {
#ifdef DEBUG
printf("h-vector fault\n");
#endif
        return 0;
        }
      break;
    case 11:
      if (h_vec[3] <= 0) {
#ifdef DEBUG
printf("h-vector fault\n");
#endif
        return 0;
        }
      break;
    case 12:
      if (h_vec[3] <= 0) {
#ifdef DEBUG
printf("h-vector fault\n");
#endif
        return 0;
        }
      break;
    case 13:
      if (h_vec[3] <= 0) {
#ifdef DEBUG
printf("h-vector fault\n");
#endif
        return 0;
        }
      break;
    case 14:
      if (h_vec[3] <= 0) {
#ifdef DEBUG
printf("h-vector fault\n");
#endif
        return 0;
        }
      break;
    case 15:
      if (h_vec[4] <= 0) {
#ifdef DEBUG
printf("h-vector fault\n");
#endif
        return 0;
        }
      break;
    }


  switch(num) {
    case 0:
      if (c_empty == 1 ||
          c_vert[facets->vert[1]] == 1 ||
          c_vert[facets->vert[2]] == 1 ||
          c_vert[facets->vert[3]] == 1 ||
          c_vert[facets->vert[4]] == 1 ||
          c_edge[facets->vert[1]][facets->vert[2]] == 1 ||
          c_edge[facets->vert[1]][facets->vert[3]] == 1 ||
          c_edge[facets->vert[1]][facets->vert[4]] == 1 ||
          c_edge[facets->vert[2]][facets->vert[3]] == 1 ||
          c_edge[facets->vert[2]][facets->vert[4]] == 1 ||
          c_edge[facets->vert[3]][facets->vert[4]] == 1 ||
          c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] == 1 ||
          c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] == 1 ||
          c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] == 1 ||
          c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] == 1
          ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
        return 0;
        }
      break;
    case 1:
      if (
          c_vert[facets->vert[1]] == 1 ||
          c_edge[facets->vert[1]][facets->vert[2]] == 1 ||
          c_edge[facets->vert[1]][facets->vert[3]] == 1 ||
          c_edge[facets->vert[1]][facets->vert[4]] == 1 ||
          c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] == 1 ||
          c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] == 1 ||
          c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] == 1
          ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
        return 0;
        }
      break;
    case 2:
      if (
          c_vert[facets->vert[2]] == 1 ||
          c_edge[facets->vert[1]][facets->vert[2]] == 1 ||
          c_edge[facets->vert[2]][facets->vert[3]] == 1 ||
          c_edge[facets->vert[2]][facets->vert[4]] == 1 ||
          c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] == 1 ||
          c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] == 1 ||
          c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] == 1 
          ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
        return 0;
        }
      break;
    case 3:
      if (
          c_vert[facets->vert[3]] == 1 ||
          c_edge[facets->vert[1]][facets->vert[3]] == 1 ||
          c_edge[facets->vert[2]][facets->vert[3]] == 1 ||
          c_edge[facets->vert[3]][facets->vert[4]] == 1 ||
          c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] == 1 ||
          c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] == 1 ||
          c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] == 1 
          ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
        return 0;
        }
      break;
    case 4:
      if (
          c_vert[facets->vert[4]] == 1 ||
          c_edge[facets->vert[1]][facets->vert[4]] == 1 ||
          c_edge[facets->vert[2]][facets->vert[4]] == 1 ||
          c_edge[facets->vert[3]][facets->vert[4]] == 1 ||
          c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] == 1 ||
          c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] == 1 ||
          c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] == 1 
          ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
        return 0;
        }
      break;
    case 5:
      if (
          c_edge[facets->vert[1]][facets->vert[2]] == 1 ||
          c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] == 1 ||
          c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] == 1
          ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
        return 0;
        }
      break;
    case 6:
      if (
          c_edge[facets->vert[1]][facets->vert[3]] == 1 ||
          c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] == 1 ||
          c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] == 1
          ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
        return 0;
        }
      break;
    case 7:
      if (
          c_edge[facets->vert[1]][facets->vert[4]] == 1 ||
          c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] == 1 ||
          c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] == 1
          ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
        return 0;
        }
      break;
    case 8:
      if (
          c_edge[facets->vert[2]][facets->vert[3]] == 1 ||
          c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] == 1 ||
          c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] == 1
          ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
        return 0;
        }
      break;
    case 9:
      if (
          c_edge[facets->vert[2]][facets->vert[4]] == 1 ||
          c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] == 1 ||
          c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] == 1
          ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
        return 0;
        }
      break;
    case 10:
      if (
          c_edge[facets->vert[3]][facets->vert[4]] == 1 ||
          c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] == 1 ||
          c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] == 1
          ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
        return 0;
        }
      break;
    case 11:
      if (
          c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] == 1 
          ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
        return 0;
        }
      break;
    case 12:
      if (
          c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] == 1 
          ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
        return 0;
        }
      break;
    case 13:
      if (
          c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] == 1 
          ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
        return 0;
        }
      break;
    case 14:
      if (
          c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] == 1 
          ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
        return 0;
        }
      break;
    case 15:
      break;
  }


  if (c_empty == 0 && m_empty == 1 && num != 0) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
    return 0;
    }
  if (c_vert[facets->vert[1]] == 0 && 
      m_vert[facets->vert[1]] == 1 &&
        !(num == 0 || num == 1) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
    return 0;
    }
  if (c_vert[facets->vert[2]] == 0 && 
      m_vert[facets->vert[2]] == 1 &&
        !(num == 0 || num == 2) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
    return 0;
    }
  if (c_vert[facets->vert[3]] == 0 && 
      m_vert[facets->vert[3]] == 1 &&
        !(num == 0 || num == 3) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
    return 0;
    }
  if (c_vert[facets->vert[4]] == 0 && 
      m_vert[facets->vert[4]] == 1 &&
        !(num == 0 || num == 4) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
    return 0;
    }
  if (c_edge[facets->vert[1]][facets->vert[2]] == 0 &&
      m_edge[facets->vert[1]][facets->vert[2]] == 1 &&
        !(num == 0 || num == 1 || num == 2 || num == 5) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
    return 0;
    }
  if (c_edge[facets->vert[1]][facets->vert[3]] == 0 &&
      m_edge[facets->vert[1]][facets->vert[3]] == 1 &&
        !(num == 0 || num == 1 || num == 3 || num == 6) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
    return 0;
    }
  if (c_edge[facets->vert[1]][facets->vert[4]] == 0 &&
      m_edge[facets->vert[1]][facets->vert[4]] == 1 &&
        !(num == 0 || num == 1 || num == 4 || num == 7) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
    return 0;
    }
  if (c_edge[facets->vert[2]][facets->vert[3]] == 0 &&
      m_edge[facets->vert[2]][facets->vert[3]] == 1 &&
        !(num == 0 || num == 2 || num == 3 || num == 8) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
    return 0;
    }
  if (c_edge[facets->vert[2]][facets->vert[4]] == 0 &&
      m_edge[facets->vert[2]][facets->vert[4]] == 1 &&
        !(num == 0 || num == 2 || num == 4 || num == 9) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
    return 0;
    }
  if (c_edge[facets->vert[3]][facets->vert[4]] == 0 &&
      m_edge[facets->vert[3]][facets->vert[4]] == 1 &&
        !(num == 0 || num == 3 || num == 4 || num == 10) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
    return 0;
    }
  if (c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] == 0 &&
      m_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] == 1 &&
        !(num == 0 || num == 1 || num == 2 || num == 3 || 
                      num == 5 || num == 6 || num == 8 || num == 11) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
    return 0;
    }
  if (c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] == 0 &&
      m_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] == 1 &&
        !(num == 0 || num == 1 || num == 2 || num == 4 || 
                      num == 5 || num == 7 || num == 9 || num == 12) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
    return 0;
    }
  if (c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] == 0 &&
      m_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] == 1 &&
        !(num == 0 || num == 1 || num == 3 || num == 4 || 
                      num == 6 || num == 7 || num == 10 || num == 13) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
    return 0;
    }
  if (c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] == 0 &&
      m_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] == 1 &&
        !(num == 0 || num == 2 || num == 3 || num == 4 || 
                      num == 8 || num == 9 || num == 10 || num == 14) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
    return 0;
    }


  return 1;

}

void putpart3(facets, num)
  Simplices* facets;
  int num;
{
  m_empty --;
  m_vert[facets->vert[1]] --;
  m_vert[facets->vert[2]] --;
  m_vert[facets->vert[3]] --;
  m_vert[facets->vert[4]] --;
  m_edge[facets->vert[1]][facets->vert[2]] --;
  m_edge[facets->vert[1]][facets->vert[3]] --;
  m_edge[facets->vert[1]][facets->vert[4]] --;
  m_edge[facets->vert[2]][facets->vert[3]] --;
  m_edge[facets->vert[2]][facets->vert[4]] --;
  m_edge[facets->vert[3]][facets->vert[4]] --;
  m_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] --;
  m_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] --;
  m_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] --;
  m_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] --;
  switch(num) {
    case 0:
      c_empty ++;
      c_vert[facets->vert[1]] ++;
      c_vert[facets->vert[2]] ++;
      c_vert[facets->vert[3]] ++;
      c_vert[facets->vert[4]] ++;
      c_edge[facets->vert[1]][facets->vert[2]] ++;
      c_edge[facets->vert[1]][facets->vert[3]] ++;
      c_edge[facets->vert[1]][facets->vert[4]] ++;
      c_edge[facets->vert[2]][facets->vert[3]] ++;
      c_edge[facets->vert[2]][facets->vert[4]] ++;
      c_edge[facets->vert[3]][facets->vert[4]] ++;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] ++;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] ++;
      c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] ++;
      c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] ++;
      h_vec[0] --;
      break;
    case 1:
      c_vert[facets->vert[1]] ++;
      c_edge[facets->vert[1]][facets->vert[2]] ++;
      c_edge[facets->vert[1]][facets->vert[3]] ++;
      c_edge[facets->vert[1]][facets->vert[4]] ++;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] ++;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] ++;
      c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] ++;
      h_vec[1] --;
      break;
    case 2:
      c_vert[facets->vert[2]] ++;
      c_edge[facets->vert[1]][facets->vert[2]] ++;
      c_edge[facets->vert[2]][facets->vert[3]] ++;
      c_edge[facets->vert[2]][facets->vert[4]] ++;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] ++;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] ++;
      c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] ++;
      h_vec[1] --;
      break;
    case 3:
      c_vert[facets->vert[3]] ++;
      c_edge[facets->vert[1]][facets->vert[3]] ++;
      c_edge[facets->vert[2]][facets->vert[3]] ++;
      c_edge[facets->vert[3]][facets->vert[4]] ++;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] ++;
      c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] ++;
      c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] ++;
      h_vec[1] --;
      break;
    case 4:
      c_vert[facets->vert[4]] ++;
      c_edge[facets->vert[1]][facets->vert[4]] ++;
      c_edge[facets->vert[2]][facets->vert[4]] ++;
      c_edge[facets->vert[3]][facets->vert[4]] ++;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] ++;
      c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] ++;
      c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] ++;
      h_vec[1] --;
      break;
    case 5:
      c_edge[facets->vert[1]][facets->vert[2]] ++;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] ++;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] ++;
      h_vec[2] --;
      break;
    case 6:
      c_edge[facets->vert[1]][facets->vert[3]] ++;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] ++;
      c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] ++;
      h_vec[2] --;
      break;
    case 7:
      c_edge[facets->vert[1]][facets->vert[4]] ++;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] ++;
      c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] ++;
      h_vec[2] --;
      break;
    case 8:
      c_edge[facets->vert[2]][facets->vert[3]] ++;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] ++;
      c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] ++;
      h_vec[2] --;
      break;
    case 9:
      c_edge[facets->vert[2]][facets->vert[4]] ++;
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] ++;
      c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] ++;
      h_vec[2] --;
      break;
    case 10:
      c_edge[facets->vert[3]][facets->vert[4]] ++;
      c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] ++;
      c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] ++;
      h_vec[2] --;
      break;
    case 11:
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] ++;
      h_vec[3] --;
      break;
    case 12:
      c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] ++;
      h_vec[3] --;
      break;
    case 13:
      c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] ++;
      h_vec[3] --;
      break;
    case 14:
      c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] ++;
      h_vec[3] --;
      break;
    case 15:
      h_vec[4] --;
      break;
    }

  return;

}

void printdata3(facets)
  Simplices* facets;
{
  Simplices* tmp_facets;
  int i;

  tmp_facets = facets;
  while(tmp_facets != NULL) {
    for (i = 1; i <= tmp_facets->vert[0] + 1; i ++)
      printf("%d ", tmp_facets->vert[i]);
    printf(": ");
    switch(tmp_facets->vert[5]) {
      case 0:
        printf("empty");
        break;
      case 1:
        printf("%d", tmp_facets->vert[1]);
        break;
      case 2:
        printf("%d", tmp_facets->vert[2]);
        break;
      case 3:
        printf("%d", tmp_facets->vert[3]);
        break;
      case 4:
        printf("%d", tmp_facets->vert[4]);
        break;
      case 5:
        printf("%d %d", tmp_facets->vert[1], tmp_facets->vert[2]);
        break;
      case 6:
        printf("%d %d", tmp_facets->vert[1], tmp_facets->vert[3]);
        break;
      case 7:
        printf("%d %d", tmp_facets->vert[1], tmp_facets->vert[4]);
        break;
      case 8:
        printf("%d %d", tmp_facets->vert[2], tmp_facets->vert[3]);
        break;
      case 9:
        printf("%d %d", tmp_facets->vert[2], tmp_facets->vert[4]);
        break;
      case 10:
        printf("%d %d", tmp_facets->vert[3], tmp_facets->vert[4]);
        break;
      case 11:
        printf("%d %d %d", tmp_facets->vert[1], tmp_facets->vert[2], tmp_facets->vert[3]);
        break;
      case 12:
        printf("%d %d %d", tmp_facets->vert[1], tmp_facets->vert[2], tmp_facets->vert[4]);
        break;
      case 13:
        printf("%d %d %d", tmp_facets->vert[1], tmp_facets->vert[3], tmp_facets->vert[4]);
        break;
      case 14:
        printf("%d %d %d", tmp_facets->vert[2], tmp_facets->vert[3], tmp_facets->vert[4]);
        break;
      case 15:
        printf("%d %d %d %d", tmp_facets->vert[1], tmp_facets->vert[2], tmp_facets->vert[3], tmp_facets->vert[4]);
        break;
      }
    printf("\n");
    tmp_facets = tmp_facets->next;
    }
  printf("\n");
  return;
}

