/*
 *
 * Function for recognizing partitionability
 * (for pure/nonpure d-dimensional complexes for any d>=1)
 *
 * by M.Hachimori, Apr 2023.
 */

#include <stdio.h>
#include <stdlib.h>
#include "s_complex.h"
#include "hash.h"


int l[MAXDIM+2], ind[MAXDIM+2];

HashCell **ht;  /* ht is the Hash table for remembering faces of the simplicial complex */

void subface_next(short subface[], int d);
void cancel_np_d(Simplices_i* facets_i);
int check_np_d(Simplices_i* facets_i);
void putpart_np_d(Simplices_i* facets_i);
void printdata_np_d(Simplices_i* facets_i);
int subface_contained(short subface1[], short subface2[], int d);
void key_encode(Simplices* facets, short subface[], short key[]);
void key_encode_i(Simplices_i* facets_i, short subface[], short key[]);


int is_partitionable_nonpure_d(facets)
  Simplices *facets;
{

  Simplices *faces, *tmp_faces;
  Simplices *tmp_facets;
  Simplices_i *facets_i, *tmp_facets_i;
  short tmp_subface[MAXDIM+2];
  int i, j, k, pow, size;
  int dim, d;
  int cnt;

  HashCell *cp;
  short key[MAXDIM+2];

  /* Check the dimension of the complex.*/
  dim = -1;
  tmp_facets = facets;
  while(tmp_facets != NULL) {
    if (tmp_facets->vert[0] > dim) dim = tmp_facets->vert[0];
    tmp_facets = tmp_facets->next;
  }

  /* initialize facets_i which records the interval partition */
  facets_i = (Simplices_i*)malloc(sizeof(Simplices_i));
  facets_i->prev = NULL;
  tmp_facets = facets;
  tmp_facets_i = facets_i;
  while(tmp_facets != NULL) {
    for (i = 0; i <= tmp_facets->vert[0]+1; i++)
      tmp_facets_i->vert[i] = tmp_facets->vert[i];
    tmp_facets_i->subface[0] = -2; /* "-2" indicates "interval is not assigned". */
    for (i = 1; i<=tmp_facets->vert[0]+1; i ++) tmp_facets_i->subface[i] = 0;
    if (tmp_facets->next != NULL) {
      tmp_facets_i->next = (Simplices_i*)malloc(sizeof(Simplices_i));
      tmp_facets_i->next->prev = tmp_facets_i;
      }
    else tmp_facets_i->next = NULL;
    tmp_facets = tmp_facets->next;
    tmp_facets_i = tmp_facets_i->next;
    }


  ht = init_hashtable();
  printf("hash table initialized\n");

  tmp_facets = facets;

  while(tmp_facets != NULL) {
    d = tmp_facets->vert[0];
    tmp_subface[0] = -1;
    for (i = 1; i <= d+2; i ++)
      tmp_subface[i] = 0;
     /* tmp_subface indicates the indices of vertices
        contained in a subface.
        When i-th vertex of a facet is contained in the subface,
        then tmp_subface[i]=1, else it is 0.
        0th entry is the dimension of the subface. 
        When processing the subsets in a sequence using 
        subset_next(), the initial subface is the empty set. 
        tmp_subface[d+2] counts the order in the sequence. */

    while(tmp_subface[0] != -2) {  /* tmp_subface[0] == -2 if 
                                       no subface is indicated. */

        key_encode(tmp_facets, tmp_subface, key);
        cp = search_cell(ht, key);
        if (cp == NULL) {
          cp = register_hash(ht, key);
          cp->mval = 1;
          }
        else cp->mval ++;    /* mval counts the multiplicity of facets containing the face. */
        cp->cval = 0;        /* cval counts the number of intervals in the partition. */

      subface_next(tmp_subface, tmp_facets->vert[0]);
      }

    tmp_facets = tmp_facets->next;
  }


/*  l[0] = 1; l[1] = 3; l[2] = 7; l[3] = 15; /* largest num of type for each dim. */
/*  ind[0] = 2; ind[1] = 3; ind[2] = 4; ind[3] = 5; /* index of the type-flag for each dim. */

  l[0] = 1;
  for (i = 1; i <= dim; i ++) 
    l[i] = (l[i-1] + 1) * 2 - 1;
  ind[0] = 2;
  for (i = 1; i <= dim; i ++)
    ind[i] = ind[i-1] + 1;


//    printdata_np_d(facets_i);

  /* main loop */
  tmp_facets_i = facets_i;
  cnt = 1;
  while(tmp_facets_i != NULL) {
#ifdef DEBUG
    printdata_np_d(facets);
#endif
    if (cnt == 0) {
      printf("current:\n");
      printdata_np_d(facets_i);
      }
    cnt = (cnt+1) % 10000000;
    /* check subfaces in the ordering -1,l,0,1,2,...,l-1 */
    d = tmp_facets_i->vert[0];
    if (tmp_facets_i->subface[ind[d]] == l[d]) {  /* backtrack */
      if (tmp_facets_i->prev == NULL) return 0;
      else {
        tmp_facets_i->subface[ind[d]] = -1;
        tmp_facets_i->subface[0] = -2;
        for(i = 1; i <= d+1; i ++) tmp_facets_i->subface[i] = 0;
        tmp_facets_i = tmp_facets_i->prev;
        cancel_np_d(tmp_facets_i);
        d = tmp_facets_i->vert[0]; /* change d to the new dim of tmp_facets_i */
        continue;
        }
      }
    else {
      subface_next(tmp_facets_i->subface, d); /* next subface */
      if (check_np_d(tmp_facets_i) == 1) {
        putpart_np_d(tmp_facets_i);
        if (tmp_facets_i->next != NULL) {
          tmp_facets_i = tmp_facets_i->next;
          continue;
          }
        else  goto PARTITIONABLE;
        }
      }
    }

  printdata_np_d(facets_i); 
  printf("error\n");
  return -1;

  PARTITIONABLE:
  printdata_np_d(facets_i); 
  return 1;

}

void subface_next(subface, d)
  short subface[];
  int d;
{
  int i, k, l;

  /* subfaces are processed in such a sequence:
     empty set -> dim d -> dim 0 -> dim 1 -> ... -> dim d-1. */
  if (subface[0] == -2){ /* if not assigned, next subset is empty set. */
    subface[0] = -1;
    for (i = 1; i <= d+1; i ++) subface[i] = 0;
    subface[d+2] = 0;
    return;
    }
  if (subface[0] == -1){ /* if empty set, next subset is dim d. */
    subface[0] = d;
    for (i = 1; i <= d+1; i ++) subface[i] = 1;
    subface[d+2] ++;
    return;
    }
  if (subface[0] == d){ /* if dim = d, next subset is dim 0. */
    subface[0] = 0;
    subface[1] = 1;
    for (i = 2; i <= d+1; i ++) subface[i] = 0;
    subface[d+2] ++;
    return;
    }
  /* otherwise, calculate the next subset */
  if(subface[d+1] == 0){
    k = d;
    while(subface[k] == 0) k --;
    subface[k] = 0;
    subface[k+1] = 1;
    subface[d+2] ++;
    return;
    }
  if(subface[d+1] == 1){
    k = d;
    while(subface[k] == 1) k --;
    if(k == d - subface[0]){  /* this is the case when it is the last subset of the dimension. */
                 /* In this case, the next subset is the first in one dim upper. */
      if (subface[0] == d-1) { /* if the dimension is d-1, then it is the last.*/
        subface[0] = -2;
        for (i = 1; i <= d+1; i ++) subface[i] = 0;
        subface[d+2] = 0;
        return;
        }
      /* if the dimiension < d-1, set to the first of the next dimension. */
      subface[0] ++;
      for (i = 1; i <= subface[0]+1; i ++)
        subface[i] = 1;
      for (i = subface[0]+2; i <= d+1; i ++)
        subface[i] = 0;
      subface[d+2] ++;
      return;
      }
    l = k;
    while(subface[l] != 1) l --;
    subface[l] = 0; subface[l+1] = 1;
    for (i = l+2; i <= l+2+d-k; i ++) subface[i] = 1;
    for (i = l+2+d-k+1; i <= d+1; i ++) subface[i] = 0;
      subface[d+2] ++;
    return;
    }
  return; /* this "return" is not needed in fact. */
}

void cancel_np_d(facets_i)
  Simplices_i* facets_i;
{
  short subface[MAXDIM+2];
  int i, d;
  HashCell* cp;
  short key[MAXDIM+1];

   /* increase m for all subfaces, decrease c for subfaces in the interval.   */
  d = facets_i->vert[0];
  subface[0] = -1;
  for (i = 1; i <= d+2; i ++)
    subface[i] = 0;  /* start from the empty set */
  while(subface[0] !=-2){
    key_encode_i(facets_i, subface, key);
    cp = search_cell(ht, key);
    cp->mval ++;
    if (subface_contained(subface, facets_i->subface, d)==1)
      cp->cval --;
    subface_next(subface, d);
  }

//  facets_i->subface[0] = -2;
//  for (i = 1; i <= d+2; i ++) facets_i->subface[i] = 0;

//printf("end: %d\n", facets_i->prev);
  return;

}

int check_np_d(facets_i)
  Simplices_i* facets_i;

{
  short subface[MAXDIM+2];
  int i, d;
  HashCell* cp;
  short key[MAXDIM+1];

  /* "multiple cover": check c for subfaces in the interval to be all 0 */
  /* "last facet forcing": c is 0 and m is 1 but the face is not in the interval */

  d = facets_i->vert[0];
  subface[0] = -1;
  for (i = 1; i <= d+2; i ++)
    subface[i] = 0;  /* start from the empty set */
  while(subface[0] !=-2){
    
    key_encode_i(facets_i, subface, key);
    cp = search_cell(ht, key);
    if (subface_contained(subface, facets_i->subface, d) == 1 && cp->cval == 1){ /* multiple cover */
      return 0; }
    if (cp->cval == 0 && cp->mval == 1 
          && subface_contained(subface, facets_i->subface, d) == 0){ /* last facet forcing */
      return 0; }
    subface_next(subface, d);
  }

  return 1;

}


void putpart_np_d(facets_i)
  Simplices_i* facets_i;
{
  short subface[MAXDIM+2];
  int i, d;
  HashCell* cp;
  short key[MAXDIM+1];

  /* decrease m for all subfaces, increase c for subfaces in the interval */
  d = facets_i->vert[0];
  subface[0] = -1;
  for (i = 1; i <= d+2; i ++)
    subface[i] = 0;  /* start from the empty set */
  while(subface[0] !=-2){
    key_encode_i(facets_i, subface, key);
    cp = search_cell(ht, key);
    cp->mval --;
    if (subface_contained(subface, facets_i->subface, d)==1){
        cp->cval = 1;
      }
    subface_next(subface, d);
  }

  return;

}

void printdata_np_d(facets_i)
  Simplices_i* facets_i;
{
  Simplices_i* tmp_facets_i;
  int i;

  tmp_facets_i = facets_i;
  while(tmp_facets_i != NULL) {
    for (i = 1; i <= tmp_facets_i->vert[0] + 1; i ++)
      printf("%d ", tmp_facets_i->vert[i]);
    printf(": ");
    if (tmp_facets_i->subface[0] == -1)
      printf("empty set");
    else if (tmp_facets_i->subface[0] >= 0) {
      for (i = 1; i <= tmp_facets_i->vert[0] + 1; i ++)
        if (tmp_facets_i->subface[i] == 1) {
          printf("%d ", tmp_facets_i->vert[i]);
          }
      }
    printf("\n");
    tmp_facets_i = tmp_facets_i->next;
    }
  printf("\n");
  return;
}


int subface_contained(subface1, subface2, d)
  short subface1[];
  short subface2[];
  int d;
{  /* return 1 if subface1 is a super-face of subface2 */
  int i;
  for (i = 1; i <= d+1; i ++) {
    if (subface1[i] < subface2[i]) return 0;
    }
  return 1;
}


void key_encode(facets, subface, key)
  Simplices* facets;
  short subface[];
  short key[];
{
  int i, idx;

  idx = 1;
  key[0] = subface[0];
  for (i = 1; i <= facets->vert[0]+1; i ++)
    if (subface[i] == 1) key[idx++] = facets->vert[i];

  return;

}

void key_encode_i(facets_i, subface, key)
  Simplices_i* facets_i;
  short subface[];
  short key[];
{
  int i, idx;

  idx = 1;
  key[0] = subface[0];
  for (i = 1; i <= facets_i->vert[0]+1; i ++)
    if (subface[i] == 1) key[idx++] = facets_i->vert[i];

  return;

}


